﻿using System.Collections.Generic;
using System;
using System.Diagnostics;
// using System.Diagnostics.Contracts;
using System.Linq;

using Buddy.CommonBot;
using Buddy.Common.Math;
using Buddy.Swtor;
using Buddy.Swtor.Objects;


namespace WingIt
{
    /// <summary>
    /// Class of helpers, because copy/pasting the same thing in every class got a bit tedious.
    /// </summary>
    /// 3/20/2012 11:21 AM Aevitas
    public static class Helpers
    {
        public delegate bool        TorCharacterPredicateDelegate(TorCharacter torCharacter);
        public delegate bool        TorEffectPredicateDelegate(TorEffect torEffect);

        /// <summary>
        /// Gets all friendly companions, including those of players around us.
        /// </summary>
        /// 3/20/2012 11:21 AM Aevitas
        public static IEnumerable<TorCharacter> Companions
        {
            get
            {
                return
                    ObjectManager.GetObjects<TorCharacter>()
                        .Where(unit => unit.Toughness == CombatToughness.Companion && unit.IsFriendly);
            }
        }

        /// <summary>
        /// Gets our current companion.
        /// </summary>
        /// 3/20/2012 11:22 AM Aevitas
        public static TorNpc Companion
        {
            get { return BuddyTor.Me.Companion; }
        }

        /// <summary>
        /// <para>Returns true, iff:<list type="bullet">
        /// <item><description><para> * ONWHOM is not null.</para></description></item>
        /// <item><description><para> * ONWHOM has MINSTACKCOUNT stacks or more of BUFFNAME</para></description></item>
        /// <item><para><description> * The buff meets any requirements imposed by EFFECTQUALIFIER.</description></para></item>
        /// </list></para>
        /// Notes:<list type="bullet">
        /// <item><description><para> * If MINSTACKCOUNT is omitted, 1 will be used.</para></description></item>
        /// <item><description><para> * If EFFECTQUALIFIER is omitted, no additional requirements will be imposed when searching for the buff.</para></description></item>
        /// </list>
        /// </summary>
        /// <param name="onWhom">must not be null.</param>
        /// <param name="buffName">if null, empty, or misspelled simply results in a 'false' returned</param>
        /// <param name="minStackCount">if unspecified, '1' will be used.</param>
        /// <param name="minTimeRemainng">if unspecified, TimeSpan.Zero will be used.</param>
        /// <returns>'true', if the toon has a minimum of the requested stacks of the identified buff.
        /// Otherwise, returns 'false'.</returns>
        /// 27Jun2012-20:17UTC chinajade
        // NB: It is tempting to make a null onWhom default to BuddyTor.Me.  However, the caller may
        // provide BuddyTor.Me.CurrentTarget as the onWhom value, and BuddyTor.Me.CurrentTarget can
        // be null on occasion (if the appropriate pre-conditions for its use have not been met). 
        // Automatically providing BuddyTor.Me for 'null' in such cases would result 
        // in the caller defaulting to the wrong TorCharacter and subsequently making bad decisions.
        // Thus, we must require onWhom to be explicitly provided for success to occur.
        public static bool      HasBuffStacks(TorCharacter                  onWhom,
                                              string                        buffName,
                                              int                           minStackCount   = 1,
                                              TorEffectPredicateDelegate    effectQualifier = null)
        {
            //Contract.Requires<ArgumentNullException>(onWhom != null);
            //Contract.Requires<ArgumentException>(!string.IsNullOrEmpty(buffName));

            effectQualifier = effectQualifier ?? (d => true);   // Resolve null effectQualifier to something sane

            return (onWhom.Buffs.FirstOrDefault(b => ((b.Name == buffName)  && effectQualifier(b) && (b.Stacks >= minStackCount))) != null);
        }

        /// <summary>
        /// <para>Returns true, iff:<list type="bullet">
        /// <item><para><description> * ONWHOM is not null.</description></para></item>
        /// <item><para><description> * ONWHOM has MINSTACKCOUNT or more stacks of DEBUFFNAME.</description></para></item>
        /// <item><para><description> * The debuff meets any requirements imposed by EFFECTQUALIFIER.</description></para></item>
        /// <item><para><description> * The toon is the caster of (owns) the DEBUFFNAME, or the DEBUFFNAME on the target is shared.</description></para></item>
        /// </list></para>
        /// Notes:<list type="bullet">
        /// <item><description><para> * If MINSTACKCOUNT is omitted, 1 will be used.</para></description></item>
        /// <item><description><para> * If EFFECTQUALIFIER is omitted, no additional requirements will be imposed when searching for the debuff.</para></description></item>
        /// <item><description><para> * You should always use this method in preference to <c>TorCharacter.HasDebuff()</c> since this one does 'ownership' checks.</para></description></item>
        /// </list>
        /// </summary>
        /// <param name="onWhom">must not be null.</param>
        /// <param name="debuffName">if null, empty, or misspelled simply results in a 'false' returned</param>
        /// <param name="minStackCount">if omitted, a value of one will be used.</param>
        /// <param name="effectQualifier">if omitted, no additional requirements will be observed when searching for the debuff.</param>
        /// <returns>'true', if the toon has a minimum of the requested stacks of the identified buff.
        /// Otherwise, returns 'false'.</returns>
        /// 27Jun2012-20:17UTC chinajade
        // NB: It is tempting to make a null onWhom default to BuddyTor.Me.  However, the caller may
        // provide BuddyTor.Me.CurrentTarget or BuddyTor.Me.Companion as the onWhom value.
        // Both of the later values can be null on occasion (if the appropriate pre-conditions
        // for their use have not been met). 
        // Automatically providing BuddyTor.Me for 'null' in such cases would result 
        // in the caller defaulting to the wrong TorCharacter and subsequently making bad decisions.
        // Thus, we must require onWhom to be explicitly provided for success to occur.
        public static bool      HasDebuffStacks(TorCharacter                onWhom,
                                                string                      debuffName,
                                                int                         minStackCount   = 1,
                                                TorEffectPredicateDelegate  effectQualifier = null)
        {
            //Contract.Requires<ArgumentNullException>(onWhom != null);
            //Contract.Requires<ArgumentException>(!string.IsNullOrEmpty(debuffName));

            effectQualifier = effectQualifier ?? (d => true);   // Resolve null effectQualifier to something sane

            return (onWhom.Debuffs.FirstOrDefault(d => ((d.Name == debuffName)
                                                        && effectQualifier(d)
                                                        && ((d.Caster.Name == BuddyTor.Me.Name) || string.IsNullOrEmpty(d.Caster.Name))    // ownership or shared
                                                        && (d.Stacks >= minStackCount)))  != null);
        }

        /// <summary>
        /// True if companion is in use by the toon.   If you use this test and it passes, you don't
        /// have to also check for Companion != null.
        /// </summary>
        public static bool  IsCompanionInUse()
        {
            return ((BuddyTor.Me.CompanionUnlocked > 0) && (Helpers.Companion != null));
        }

        /// <summary>
        /// Returns true, iff:<list type="bullet">
        /// <item><para><description> * TARGET exists.  (no checks for 'hostile', 'dead', etc).</description></para></item>
        /// <item><para><description> * TARGET's distance from the toon is between MINRANGE and MAXRANGE, inclusive.</description></para></item>
        /// </list>Notes:<list>
        /// <item><description><para> * MINRANGE and MAXRANGE must be non-negative.</para></description></item>
        /// <item><description><para> * If TARGET is omitted, it defaults to <c>BuddyTor.Me.CurrentTarget</c>.
        /// If BuddyTor.Me.CurrentTarget is null, 'false' is returned.</para></description></item>
        /// </list>
        /// </summary>
        /// <param name="maxRange">must be zero or positive. If omitted, a value of float.MaxValue is used.</param>
        /// <param name="minRange">must be zero or positive. If omitted, a value of zero is used.</param>
        /// <param name="target">if omitted, defaults to BuddyTor.Me.CurrentTarget</param>
        /// <returns>'true', if current target distance is greater than or equal to minRange, and less than or equal to maxRange.
        /// Otherwise, returns 'false'.</returns>
        public static bool      IsInRange(float         maxRange,
                                          float         minRange   = 0.0f,
                                          TorCharacter  target     = null)
        {
            // Distances should never be negative...
            // TODO: Fixed until figure out why this post: http://www.thebuddyforum.com/buddy-wing-forum/buddy-wing-classes/52113-wingit-all-one-26.html#post601427
            //Contract.Requires<ArgumentException>(minRange >= 0.0f);
            //Contract.Requires<ArgumentException>(maxRange >= 0.0f);
            //Contract.Requires<ArgumentException>(minRange <= maxRange);

            target = target ?? BuddyTor.Me.CurrentTarget;       // Resolve null target into something sane

            return ((target != null) && (target.Distance >= minRange) && (target.Distance <= maxRange));
        }

        /// <summary>
        /// Gets the enemies currently attacking us.
        /// </summary>
        /// 3/20/2012 11:23 AM Aevitas
        public static IEnumerable<TorCharacter> Targets
        {
            // We have a defensive null coalescing operator here, as BWcore is returning 'null' in some places
            // where it should be returning empty lists.  If BWcore behaves, then the right-hand side of the
            // operator (and its associated expense) will never be executed.
            // 27Jun2012-21:43UTC chinajade
            get { return (BuddyTor.Me.EnemiesAttackers ?? new List<TorCharacter>()); }
        }

        /// <summary>
        /// Gets the weakest target, that is the one with lowest actual HP.
        /// </summary>
        /// 3/20/2012 12:11 PM Aevitas
        public static TorCharacter WeakestTarget { get { return Targets.OrderBy(t => t.Health).FirstOrDefault(); } }

        /// <summary>
        /// Gets the players in our group.
        /// </summary>
        /// 3/22/2012 4:02 PM Aevitas
        public static IEnumerable<TorPlayer> Group
        {
            get { return ObjectManager.GetObjects<TorPlayer>().Where(p => p.GroupId == BuddyTor.Me.GroupId); }
        }

        /// <summary>
        /// <para>Convenience wrapper around <c>MobsAround(TorCharacter target,
        /// float distance, TorCharacterPredicateDelegate mobQualifier)</c>, where</para>
        /// <para>TARGET is supplied as <c>BuddyTor.Me</c>.</para>
        /// </summary>
        /// <param name="distance">Maximum distance from the player in which an Npc should be counted.</param>
        /// <returns></returns>
        /// Neo93
        public static int   MobsAround(float distance, TorCharacterPredicateDelegate mobQualifier = null)
        {
            return (MobsAround(BuddyTor.Me, distance, mobQualifier));
        }

        /// <summary>
        /// <para>Counts the non-friendly, non-dead Npcs around the provided TARGET within the given DISTANCE.</para>
        /// <para>An Npc must meet the optional MOBQUALIFIER to be counted.</para>
        /// <para>Notes:<list type="bullet">
        /// <item><description><para> * If MOBQUALIFIER is omitted, all non-friendly, non-dead mobs in the area will be counted.</para></description></item>
        /// <item><description><para> * To count only hostile mobs (versus non-friendly), supply t.IsHostile as a MOBQUALIFIER.</para></description></item>
        /// <item><description><para> * The MOBQUALIFIER is frequently used to exclude stunned mobs, or mobs with a certain debuff from the count.</para></description></item>
        /// <item><description><para> * The area tested is that of a 'filled cicle' around the TARGET.</para></description></item>
        /// </list>
        /// </para>
        /// </summary>
        /// <param name="originMob">must not be null.</param>
        /// <param name="distance">Maximum distance from the player in which an Npc should be counted.</param>
        /// <param name="mobQualifier">if omitted, no additional requirements will be imposed on a mob to have it counted.</param>
        /// <returns></returns>
        public static int   MobsAround(TorCharacter                     originMob,
                                       float                            distance,
                                       TorCharacterPredicateDelegate    mobQualifier    = null)
        {
            //Contract.Requires<ArgumentNullException>(originMob != null);
            //Contract.Ensures(Contract.Result<int>() >= 0);

            var     targets     = ObjectManager.GetObjects<Buddy.Swtor.Objects.TorNpc>();

            mobQualifier = mobQualifier ?? (m => true);     // Resolve null mobQualifier to something sane

            // Note we use a !IsFriendly test instead of an IsHostile test such that 'neutral' targets will be
            // appropriately counted. Since any cleaving or AoE type attack will affect neutral mobs, this is a must for a default.
            // We also have a defensive check for 'targets == null' here, because the BWcore is returning 'null'
            // instead of empty lists in some circumstances.
            // 27Jun2012-21:44UTC chinajade
            return ((targets == null)
                    ? 0
                    : targets.Count(t => (Vector3.Distance(originMob.Position, t.Position) <= distance) &&
                                          mobQualifier(t) &&
                                          !t.IsDead &&
                                          !t.IsFriendly));
        }

        /// <summary>
        /// <para>Counts the non-friendly, non-dead Npcs around the player that can be affected by using ABILITYNAME.</para>
        /// <para>Notes:<list type="bullet">
        /// <item><description><para> * The area tested may be 'donut shaped' around the toon.
        /// This happens when the ability has a minimum range at which it can be used.</para></description></item>
        /// </list>
        /// </para>
        /// </summary>
        /// <param name="abilityName"></param>
        /// <returns></returns>
        ///  28Jun2012-17:18UTC chinajade
        public static int   MobsWithinAbilityRange(string   abilityName)
        {
            //Contract.Requires<ArgumentNullException>(!string.IsNullOrEmpty(abilityName));
            //Contract.Ensures(Contract.Result<int>() >= 0);

            TorAbility      ability     = AbilityManager.KnownAbilities.FirstOrDefault(a => (a.Name == abilityName));
            var             targets     = ObjectManager.GetObjects<Buddy.Swtor.Objects.TorNpc>();

            // Note we use a !IsFriendly test instead of an IsHostile test such that 'neutral' targets will be
            // appropriately counted. Since any cleaving or AoE type attack will affect neutral mobs, this is a must.
            return (((ability == null) || (targets == null))
                    ? 0
                    : targets.Count(t => !t.IsDead && !t.IsFriendly && IsInRange(ability.MaxRange, ability.MinRange, t)));
        }


        /// <summary>
        /// Gets the lowest health player.
        /// </summary>
        /// 3/22/2012 4:06 PM Aevitas
        public static TorCharacter LowestHealthPlayer
        {
            get
            {
                return
                    ObjectManager.GetObjects<TorCharacter>()
                        .OrderBy(p => p.HealthPercent)
                        .FirstOrDefault(p => p.IsFriendly && !p.IsDead && p.InLineOfSight);
            }
        }

        /// <summary>
        /// Gets the lowest health player in group.
        /// </summary>
        /// Neo93
        public static TorPlayer LowestHealthGroupPlayer
        {
            get
            {
                return 
                    Group.OrderBy(p => p.HealthPercent)
                        .FirstOrDefault(p => p.IsFriendly && !p.IsDead && p.InLineOfSight);
            }
        }

        /// <summary>
        /// Gets the players in the group with >= 1 debuff.
        /// </summary>
        /// Neo93
        public static TorPlayer DebuffTarget
        {
            get { return Group.OrderByDescending(o => o.Debuffs).FirstOrDefault(c => c.Debuffs.Count() >= 1); }
        }

        /// <summary>
        /// Gets all players.
        /// </summary>
        /// 3/22/2012 4:06 PM Aevitas
        public static IEnumerable<TorPlayer> Players { get { return ObjectManager.GetObjects<TorPlayer>(); } }

        /// <summary>
        /// Gets all units.
        /// </summary>
        /// 3/22/2012 4:07 PM Aevitas
        public static IEnumerable<TorCharacter> Units { get { return ObjectManager.GetObjects<TorCharacter>(); } }
        
        /// <summary>
        /// Gets the central target of specified group.
        /// </summary>
        /// <param name="group">The group.</param>
        /// <returns></returns>
        /// 3/23/2012 8:09 PM Aevitas
        public static TorCharacter CentralTarget(List<TorCharacter> group)
        {
            // Handle single targets without going through too much difficult stuff.
            if (group.Count() == 1)
                return group.FirstOrDefault();

            TorCharacter ret = null;

            float distance = float.MaxValue;

            // We're going to take a look at the cumulative distance of all members in the group in relation
            // to the unit we're evaluating. The one most in the center will have the smallest average distance to the others.
            foreach (var unit in group)
            {
                var totalDistance = group.Sum(t => t.Distance);
                var averageDistance = totalDistance/group.Count;

                if (averageDistance < distance)
                {
                    ret = unit;
                    distance = averageDistance;
                }
            }

            return ret;
        }

        /// <summary>
        /// Returns the central target of a group, based on who is closest to the "central point" of a group.
        /// </summary>
        /// <param name="group">The group.</param>
        /// <returns></returns>
        /// 3/24/2012 2:55 PM Aevitas
        public static TorCharacter CentralTargetEx(List<TorCharacter> group)
        {
            TorCharacter ret = null;

            if (group.Count == 1)
                return group.FirstOrDefault();

            Vector3 totalPosition = new Vector3(0, 0, 0);
            totalPosition = @group.Aggregate(totalPosition, (current, unit) => current + unit.Position);

            Vector3 centralPoint = totalPosition/group.Count;
            float distanceToCentral = float.MaxValue;

            foreach (var unit in group)
            {
                var distance = Vector3.Distance(unit.Position, centralPoint);
                if (distance < distanceToCentral)
                {
                    ret = unit;
                    distanceToCentral = distance;
                }
            }

            return ret;
        }

        /// <summary>
        /// Returns the current Tank
        /// </summary>
        /// Neo93
        public static TorCharacter Tank
        {
            get
            {
                if (Group.FirstOrDefault(p => tankSpecs.Contains(SpecHandler.GetSpec(p))) != null)
                {
                    return Group.FirstOrDefault(p => tankSpecs.Contains(SpecHandler.GetSpec(p)));
                }
                else if (IsCompanionInUse())
                    { return Companion; }

                return (BuddyTor.Me); 
            }
        }

        private static SkillTreeId[] tankSpecs = { SkillTreeId.AssassinDarkness, SkillTreeId.GuardianDefense,
                                                   SkillTreeId.JuggernautImmortal, SkillTreeId.PowertechShieldTech,
                                                   SkillTreeId.ShadowCombat, SkillTreeId.VanguardShieldSpecialist,
                                                 };

        public static TorCharacter ccTarget()
        {
            var previousTarget = BuddyTor.Me.CurrentTarget;
            return
            Targets.FirstOrDefault(
                t =>
                t != previousTarget && (t.Toughness == CombatToughness.Player || t.Toughness == CombatToughness.Strong || t.Toughness == CombatToughness.Standard || t.Toughness == CombatToughness.Weak) && !t.IsDead) ??
           Targets.FirstOrDefault(t => t != previousTarget);
        }
    }
}
