﻿using System;
using System.Collections.Generic;
using System.Diagnostics.Contracts;
using System.Linq;
using System.Linq.Expressions;
using System.Reflection;
using System.Text;

using Buddy.BehaviorTree;
using Buddy.Common;
using Buddy.Common.Math;
using Buddy.CommonBot;
using Buddy.Swtor;

using Action = Buddy.BehaviorTree.Action;


namespace WingIt.Routines.Mirrors
{
    public class Mirror
    {
        #region Cast methods
        /// <summary>
        /// <para>Casts ABILITYNAME, if REQUIREMENTS are met.  ABILITYNAME can be supplied as the toon's class ability, or the toon's mirror-class ability--
        /// this method will figure out the correct one to use.</para>
        /// <para>For use in a behavior tree <c>PrioritySelector</c>.
        /// Returns <c>RunStatus.Success</c> if the ABILITYNAME was actually cast.  Otherwise, returns <c>RunStatus.Failure</c></para>
        /// <para>Failure to cast the spell can occur for a number of reasons:<list type="bullet">
        /// <item><description><para> * the REQUIREMENTS are not met</para></description></item> 
        /// <item><description><para> * the toon has not learned the ability (or the ABILITYNAME is misspelled by the caller)</para></description></item> 
        /// <item><description><para> * the target is not within the spell's range</para></description></item> 
        /// <item><description><para> * the ability is on cooldown</para></description></item>
        /// </list></para>
        /// <para>Notes:<list type="bullet">
        /// <item><description><para> * If ABILITYNAME does not belong to the toon's class, or the toon's mirror-class, a MaintenanceError is logged.</para></description></item>
        /// <item><description><para> * If REQUIREMENTS are omitted, ability use will be blindly attempted.</para></description></item> 
        /// <item><description><para> * The caller does not need to provide REQUIREMENTS for ability learned, distance to use, or cooldowns.
        /// All these checks are implicitly built into this method.</para></description></item> 
        /// </list>
        /// <item><description><para>Prefer this method when possible since this it does class-mirror substitutions.</para></description></item>
        /// </para>
        /// </summary>
        /// <param name="abilityName"></param>
        /// <param name="requirements"></param>
        /// <returns></returns>
        public static Composite Cast(string abilityName, BooleanValueDelegate requirements = null)
        {
            Contract.Requires<ArgumentException>(!string.IsNullOrEmpty(abilityName));

            requirements = requirements ?? (ctx => true);       // Resolve null requirements to something sane

            // CombatRoutine.Cast() checks that we have learned the ability and
            // that the target is in range, so no need to duplicate that here...
            // (Verified by a conversation with Aevitas while he was looking at the BWcore source -- 28Jun2012-19:07UTC chinajade)
            return (WingIt.Instance.Cast(LookupAbility(abilityName), requirements));
        }

        /// <summary>
        /// <para>Casts ABILITYNAME on ONCHARACTER, if REQUIREMENTS are met.  ABILITYNAME can be supplied as the toon's class ability, or the toon's mirror-class ability--
        /// this method will figure out the correct one to use.</para>
        /// <para>For use in a behavior tree <c>PrioritySelector</c>.
        /// Returns <c>RunStatus.Success</c> if the ABILITYNAME was actually cast.  Otherwise, returns <c>RunStatus.Failure</c></para>
        /// <para>Failure to cast the spell can occur for a number of reasons:<list type="bullet">
        /// <item><description><para> * the REQUIREMENTS are not met</para></description></item> 
        /// <item><description><para> * the toon has not learned the ability (or the ABILITYNAME is misspelled by the caller)</para></description></item> 
        /// <item><description><para> * the target is not within the spell's range</para></description></item> 
        /// <item><description><para> * the ability is on cooldown</para></description></item>
        /// </list></para>
        /// <para>Notes:<list type="bullet">
        /// <item><description><para> * If ABILITYNAME does not belong to the toon's class, or the toon's mirror-class, a MaintenanceError is logged.</para></description></item>
        /// <item><description><para> * If REQUIREMENTS are omitted, ability use will be blindly attempted.</para></description></item> 
        /// <item><description><para> * The caller does not need to provide REQUIREMENTS for ability learned, distance to use, or cooldowns.
        /// All these checks are implicitly built into this method.</para></description></item> 
        /// </list>
        /// <item><description><para>Prefer this method when possible since this it does class-mirror substitutions.</para></description></item>
        /// </para>
        /// </summary>
        /// <param name="abilityName"></param>
        /// <param name="requirements"></param>
        /// <returns></returns>
        public static Composite Cast(string abilityName, CharacterSelectionDelegate onCharacter, BooleanValueDelegate requirements = null)
        {
            Contract.Requires<ArgumentException>(!string.IsNullOrEmpty(abilityName));

            requirements = requirements ?? (ctx => true);       // Resolve null requirements to something sane

            // CombatRoutine.Cast() checks that we have learned the ability and
            // that the target is in range, so no need to duplicate that here...
            // (Verified by a conversation with Aevitas while he was looking at the BWcore source -- 28Jun2012-19:07UTC chinajade)
            return WingIt.Instance.Cast(LookupAbility(abilityName), onCharacter, requirements);
        }

        /// <summary>
        /// <para>Casts ABILITYNAME on LOCATION, if REQUIREMENTS are met.  ABILITYNAME can be supplied as the toon's class ability, or the toon's mirror-class ability--
        /// this method will figure out the correct one to use..</para>
        /// <para>For use in a behavior tree <c>PrioritySelector</c>.</para>
        /// <para>Returns <c>RunStatus.Success</c> if the ABILITYNAME was actually cast.  Otherwise, returns <c>RunStatus.Failure</c></para>
        /// <para>Failure to cast the spell can occur for a number of reasons:<list type="bullet">
        /// <item><description><para> * the REQUIREMENTS are not met</para></description></item> 
        /// <item><description><para> * the toon has not learned the ability (or the ABILITYNAME is misspelled by the caller)</para></description></item> 
        /// <item><description><para> * the target location is not within the spell's range</para></description></item> 
        /// <item><description><para> * the ability is on cooldown</para></description></item> 
        /// </list></para>
        /// <para>Notes:<list type="bullet">
        /// <item><description><para> * If ABILITYNAME does not belong to the toon's class, or the toon's mirror-class, a MaintenanceError is logged.</para></description></item>
        /// <item><description><para> * If REQUIREMENTS are omitted, ability use will be blindly attempted.</para></description></item> 
        /// <item><description><para> * The caller does not need to provide REQUIREMENTS for ability learned, distance to use, or cooldowns.
        /// All these checks are implicitly built into this method.</para></description></item> 
        /// </list>
        /// <item><description><para>Prefer this method when possible since this it does class-mirror substitutions.</para></description></item>
        /// </para>
        /// </summary>
        /// <param name="abilityName"></param>
        /// <param name="requirements"></param>
        /// <returns></returns>
        public static Composite CastOnGround(string abilityName, BooleanValueDelegate requirements, CommonBehaviors.Retrieval<Buddy.Common.Math.Vector3> location)
        {
            Contract.Requires<ArgumentException>(!string.IsNullOrEmpty(abilityName));

            requirements = requirements ?? (ctx => true);   // Resolve a null argument into something sane

            return new Decorator(doWhen =>  requirements(doWhen) && (location != null) && (location(doWhen) != Vector3.Zero),
                new Action(ctx => AbilityManager.Cast(LookupAbility(abilityName), location(ctx))));
        }
        #endregion


        // NB: HasMirrorBuff() and HasMirrorDebuff() methods are in MirrorExtensions.cs.


        #region Lookup methods
        /// <summary>
        /// <para>Given SRCABILITYNAME, returns the correct ability name for the class.</para>
        /// <para>Notes:<list type="bullet">
        /// <item><description><para> * SRCABILITYNAME may be the toon's class ability name, or the mirror-class ability name.</para></description></item>
        /// <item><description><para> * If the ability name cannot be located, a MaintenanceError is logged, and an empty string returned.
        /// This is usually due to spelling errors by the caller.</para></description></item>
        /// <item><description><para> * Prefer using a <c>Mirror.Cast()</c> method variant rather looking up a abiilty name explicitly with this method.</para></description></item>
        /// </list>
        /// </para>
        /// </summary>
        /// <param name="srcAbilityName"></param>
        /// <returns></returns>
        public static string    LookupAbility(string    srcAbilityName)
        {
            return (Lookup(MirrorData.MasterMirror_AbilityMap, srcAbilityName, "Ability"));
        }

        /// <summary>
        /// <para>Given SRCBUFFNAME, returns the correct buff name for the class.</para>
        /// <para>Notes:<list type="bullet">
        /// <item><description><para> * SRCBUFFNAME may be the toon's class buff name, or the mirror-class buff name.</para></description></item>
        /// <item><description><para> * If the buff name cannot be located, a MaintenanceError is logged, and an empty string returned.
        /// This is usually due to spelling errors by the caller.</para></description></item>
        /// <item><description><para> * Prefer using the <c>TorCharacter.HasMirrorBuff()</c> extension method rather looking up a buff name explicitly with this method.</para></description></item>
        /// </list>
        /// </para>
        /// </summary>
        /// <param name="srcBuffName"></param>
        /// <returns></returns>
        public static string    LookupBuff(string   srcBuffName)
        {
            return (Lookup(MirrorData.MasterMirror_BuffMap, srcBuffName, "Buff"));
        }

        /// <summary>
        /// <para>Given SRCDEBUFFNAME, returns the correct buff name for the class.</para>
        /// <para>Notes:<list type="bullet">
        /// <item><description><para> * SRCDEBUFFNAME may be the toon's class debuff name, or the mirror-class debuff name.</para></description></item>
        /// <item><description><para> * If the debuff name cannot be located, a MaintenanceError is logged, and an empty string returned.
        /// This is usually due to spelling errors by the caller.</para></description></item>
        /// <item><description><para> * Prefer using the <c>TorCharacter.HasMirrorDebuff()</c> extension method rather looking up a debuff name explicitly with this method.</para></description></item>
        /// </list>
        /// </para>
        /// </summary>
        /// <param name="srcDebuffName"></param>
        /// <returns></returns>
        public static string    LookupDebuff(string srcDebuffName)
        {
            return (Lookup(MirrorData.MasterMirror_DebuffMap, srcDebuffName, "Debuff"));
        }

        private static string    Lookup(Dictionary<CharacterClass, Dictionary<string, string>>  masterMirrorMap,
                                        string                                                  srcName,
                                        string                                                  pairTypeName)
        {
            Contract.Requires<ArgumentException>(masterMirrorMap != null);
            Contract.Requires<ArgumentException>(!string.IsNullOrEmpty(srcName));
            Contract.Requires<ArgumentException>(!string.IsNullOrEmpty(pairTypeName));
            Contract.Ensures(Contract.Result<string>() != null);

            string          destName;
            CharacterClass  normalizedClass = GetNormalizedClass();

            Dictionary<string, string>  mirrorAbilitiesForward = masterMirrorMap[normalizedClass];
            if (mirrorAbilitiesForward.TryGetValue(srcName, out destName))
                { return (srcName); }

            Dictionary<string, string>  mirrorAbilitiesReverse = masterMirrorMap[GetMirrorClass(normalizedClass)];
            if (mirrorAbilitiesReverse.TryGetValue(srcName, out destName))
                { return (destName); }

            Logger.MaintenanceError("{0} '{2}' is neither a {3} nor a {4} {1}. Spelling error in {0} name?",
                                    pairTypeName, pairTypeName.ToLower(),
                                    srcName, normalizedClass, GetMirrorClass(normalizedClass));
            return (string.Empty);
        }
        #endregion  // Lookup methods


        #region Utility methods
        /// <summary>
        /// <para>Given CHARACTERCLASS, this method returns the 'mirror' class.</para>
        /// <para>Notes:<list type="bullet">
        /// <item><description><para> * CHARACTERCLASS may be either a basic class or an advanced class.
        /// The appropriate mirror class is returned in either case.</para></description></item>
        /// <item><description><para> * The mirror class of CharacterClass.Unknown is CharacterClass.Unknown.</para></description></item>
        /// </list></para>
        /// </summary>
        /// <param name="characterClass"></param>
        /// <returns></returns>
        public static CharacterClass    GetMirrorClass(CharacterClass   characterClass)
        {
            CharacterClass  mirrorClass;

            if (MirrorData.Class_EmpireToRepublic.TryGetValue(characterClass, out mirrorClass))
                { return (mirrorClass); }

            if (MirrorData.Class_RepublicToEmpire.TryGetValue(characterClass, out mirrorClass))
                { return (mirrorClass); }

            Logger.MaintenanceError("Character class '{0}' is neither Empire nor Republic", characterClass);
            return (CharacterClass.Unknown);
        }

        /// <summary>
        /// <para>Returns the advanced class for <c>BuddyTor.Me</c>, if one exists.  If the toon has yet to acquire
        /// an advanced class, the basic class is returned.</para>
        /// <para>Notes:<list type="bullet">
        /// <item><description><para> * If an advanced class is returned, it is returned as a <c>CharacterClass</c> enumeration
        /// value--not an <c>AdvancedClass</c> enumeration.  This is what is meant by 'normalizing'.</para></description></item>
        /// </list>
        /// </para>
        /// </summary>
        /// <returns></returns>
        public static CharacterClass    GetNormalizedClass()
        {
            if (BuddyTor.Me.AdvancedClass == AdvancedClass.None)
                { return (BuddyTor.Me.Class); }
            
            return (MirrorData.AdvancedToCharacterClass[BuddyTor.Me.AdvancedClass]);
        }

        /// <summary>
        /// <para>Returns true, if the CHARACTERCLASS is associated with the Empire.</para>
        /// </summary>
        /// <param name="characterClass"></param>
        /// <returns></returns>
        public static bool      IsEmpireClass(CharacterClass    characterClass)
        {
            CharacterClass  dummy;

            return (MirrorData.Class_EmpireToRepublic.TryGetValue(characterClass, out dummy));
        }

        /// <summary>
        /// Returns true, if the CHARACTERCLASS is associated with the Republic.
        /// </summary>
        /// <param name="characterClass"></param>
        /// <returns></returns>
        public static bool      IsRepublicClass(CharacterClass  characterClass)
        {
            CharacterClass  dummy;

            return (MirrorData.Class_RepublicToEmpire.TryGetValue(characterClass, out dummy));
        }
        #endregion  // Utility methods


        #region Validation
        /// <summary>
        /// Flags any maintenance errors that may be introduced into the master mapping tables.
        /// </summary>
        /// <param name="abilityPairs"></param>
        /// <param name="allowedEmpireClasses"></param>
        /// <param name="allowedRepublicClasses"></param>
        /// <returns></returns>
        public static bool      ValidateTableMaintenance(MirrorData.AbilityPair[]   abilityPairs,
                                                         string                     varName_abilityPairs,
                                                         Dictionary<CharacterClass, Dictionary<string, string>>  masterMirrorMap,
                                                         string                     varName_masterMirrorMap,
                                                         CharacterClass[]           allowedEmpireClasses,
                                                         CharacterClass[]           allowedRepublicClasses)
        {
            Contract.Requires<ArgumentNullException>(abilityPairs != null);
            Contract.Requires<ArgumentException>(!string.IsNullOrEmpty(varName_abilityPairs));
            Contract.Requires<ArgumentNullException>(masterMirrorMap != null);
            Contract.Requires<ArgumentException>(!string.IsNullOrEmpty(varName_masterMirrorMap));
            Contract.Requires<ArgumentNullException>(allowedEmpireClasses != null);
            Contract.Requires<ArgumentException>(allowedEmpireClasses.Length > 0);
            Contract.Requires<ArgumentNullException>(allowedRepublicClasses != null);
            Contract.Requires<ArgumentException>(allowedRepublicClasses.Length > 0);

            IEnumerable<CharacterClass>     allowedClasses  = allowedEmpireClasses.Concat(allowedRepublicClasses);
            Dictionary<string, string>      dummy;
            bool                            isValid         = true;

            // Validate allowed classes --
            foreach (var characterClass in allowedEmpireClasses)
            {
                if (!IsEmpireClass(characterClass))
                {
                    Logger.MaintenanceError("Argument 'allowedEmpireClasses' has class '{0}' that is not Empire faction",
                                            characterClass);
                    isValid = false;
                }
            }

            foreach (var characterClass in allowedRepublicClasses)
            {
                if (!IsRepublicClass(characterClass))
                {
                    Logger.MaintenanceError("Argument 'allowedRepublicClasses' has class '{0}' that is not Republic faction",
                                            characterClass);
                    isValid = false;
                }
            }

            // Make certain forward and reverse maps are present --
            foreach (var characterClass in allowedClasses)
            {
                if (!masterMirrorMap.TryGetValue(characterClass, out dummy))
                {
                    Logger.MaintenanceError("MirrorAbilityMap '{0}' is missing an entry for the '{1}' key.",
                                            varName_abilityPairs, characterClass);
                    isValid = false;
                }
            }

            // Examine each pair entry --
            foreach (var entry in abilityPairs)
            {
                // Empire ability name should not be empty...
                if (string.IsNullOrEmpty(entry.Empire.AbilityName))
                {
                    Logger.MaintenanceError("In {0} for entry {1}...\n"
                                             + "    Empire ability name may not be empty.",
                                             varName_masterMirrorMap, entry.ToString());
                    isValid = false;
                }

                // Empire class needs to be one of those allowed...
                if ((allowedEmpireClasses != null) && !allowedEmpireClasses.Contains(entry.Empire.NormalizedClass))
                {
                    Logger.MaintenanceError("In {0} for entry {1}...\n"
                                             + "    Class '{2}' is not allowed for Empire Class in the table.\n"
                                             + "    Allowed classes: {3}.",
                                             varName_masterMirrorMap, entry.ToString(), entry.Empire.NormalizedClass,
                                             string.Join(", ", allowedEmpireClasses));
                    isValid = false;
                }

                // Republic ability name should not be empty...
                if (string.IsNullOrEmpty(entry.Republic.AbilityName))
                {
                    Logger.MaintenanceError("In {0} for entry {1}...\n"
                                             + "    Republic ability name may not be empty.",
                                             varName_masterMirrorMap, entry.ToString());
                    isValid = false;
                }

                // Republic class needs to be one of those allowed...
                if ((allowedRepublicClasses != null) && !allowedRepublicClasses.Contains(entry.Republic.NormalizedClass))
                {
                    Logger.MaintenanceError("In {0} for entry {1}...\n"
                                             + "    Class '{2}' is not allowed for Republic Class in the table.\n"
                                             + "    Allowed classes: {3}.",
                                             varName_masterMirrorMap, entry.ToString(), entry.Republic.NormalizedClass,
                                             string.Join(", ", allowedRepublicClasses));
                    isValid = false;
                }

                // Republic & Empire classes should be mirrors of each other...
                if (GetMirrorClass(entry.Empire.NormalizedClass) != entry.Republic.NormalizedClass)
                {
                    Logger.MaintenanceError("In {0} for entry {1}...\n"
                                             + "    Class '{2}' is the mirror of class '{3}' (not the provided '{4}').",
                                             varName_masterMirrorMap, entry.ToString(),
                                             GetMirrorClass(entry.Empire.NormalizedClass), entry.Empire.NormalizedClass,
                                             entry.Republic.NormalizedClass);
                    isValid = false;
                }
            }

            return (isValid);
        }
        #endregion  // Validation

    }
}
