﻿using System;
using System.Collections.Generic;
using System.Diagnostics.Contracts;
using System.Linq;
using System.Text;

using Buddy.BehaviorTree;
using Buddy.Common;
using Buddy.Common.Math;
using Buddy.CommonBot;
using Buddy.Swtor;
using Buddy.Swtor.Objects;


namespace WingIt.Routines.Mirrors
{
    public static class MirrorExtensions
    {
        public delegate bool        TorEffectPredicateDelegate(TorEffect torEffect);

        /// <summary>
        /// <para>Returns true, iff:<list type="bullet">
        /// <item><para><description> * MINSTACKCOUNT or more stacks of BUFFNAME are found.</description></para></item>
        /// <item><para><description> * The buff meets any requirements imposed by EFFECTQUALIFIER.</description></para></item>
        /// </list></para>
        /// Notes:<list type="bullet">
        /// <item><description><para> * The BUFFNAME can either be the toon's class buff name, or the toon's mirror-class buff name.
        /// This method will figure out the correct one to use.</para></description></item>
        /// <item><description><para> * If MINSTACKCOUNT is omitted, 1 will be used.</para></description></item>
        /// <item><description><para> * If EFFECTQUALIFIER is omitted, no additional requirements will be imposed when searching for the buff.</para></description></item>
        /// <item><description><para>Prefer this method to <c>TorCharacter.HasBuff()</c> since this one does class-mirror substitutions.</para></description></item>
        /// </list>
        /// </summary>
        /// <param name="onWhom">must not be null.</param>
        /// <param name="debuffName">if null, empty, or misspelled simply results in a 'false' returned</param>
        /// <param name="minStackCount">if omitted, a value of one will be used.</param>
        /// <param name="effectQualifier">if omitted, no additional requirements will be observed when searching for the debuff.</param>
        /// <returns>'true', if the toon has a minimum of the requested stacks of the identified buff.
        /// Otherwise, returns 'false'.</returns>
        /// 29Jun2012-14:08UTC chinajade
        public static bool      HasMirrorBuff(this TorCharacter             onWhom,
                                              string                        buffName,
                                              int                           minStackCount   = 1,
                                              TorEffectPredicateDelegate    effectQualifier = null)
        {
            Contract.Requires<ArgumentNullException>(onWhom != null);
            Contract.Requires<ArgumentException>(!string.IsNullOrEmpty(buffName));

            effectQualifier = effectQualifier ?? (d => true);   // Resolve null effectQualifier to something sane

            string      normalizedBuffName = Mirror.LookupBuff(buffName);

            return (onWhom.Buffs.FirstOrDefault(b => ((b.Name == normalizedBuffName)
                                                            && effectQualifier(b)
                                                            && (b.Stacks >= minStackCount))) != null);
        }

        /// <summary>
        /// <para>Returns true, iff:<list type="bullet">
        /// <item><para><description> * MINSTACKCOUNT or more stacks of DEBUFFNAME are found.</description></para></item>
        /// <item><para><description> * The debuff meets any requirements imposed by EFFECTQUALIFIER.</description></para></item>
        /// <item><para><description> * The toon is the caster of (owns) the DEBUFFNAME, or the DEBUFFNAME has shared ownership.</description></para></item>
        /// </list></para>
        /// Notes:<list type="bullet">
        /// <item><description><para> * The DEBUFFNAME can either be the toon's class debuff name, or the toon's mirror-class debuff name.
        /// This method will figure out the correct one to use.</para></description></item>
        /// <item><description><para> * If MINSTACKCOUNT is omitted, 1 will be used.</para></description></item>
        /// <item><description><para> * If EFFECTQUALIFIER is omitted, no additional requirements will be imposed when searching for the debuff.</para></description></item>
        /// <item><description><para>Prefer this method to <c>TorCharacter.HasDebuff()</c> since this one does 'ownership' checks and class-mirror substitutions.</para></description></item>
        /// </list>
        /// </summary>
        /// <param name="onWhom">must not be null.</param>
        /// <param name="debuffName">if null, empty, or misspelled simply results in a 'false' returned</param>
        /// <param name="minStackCount">if omitted, a value of one will be used.</param>
        /// <param name="effectQualifier">if omitted, no additional requirements will be observed when searching for the debuff.</param>
        /// <returns>'true', if the toon has a minimum of the requested stacks of the identified buff.
        /// Otherwise, returns 'false'.</returns>
        /// 29Jun2012-14:08UTC chinajade
        public static bool      HasMirrorDebuff(this TorCharacter           onWhom,
                                                string                      debuffName,
                                                int                         minStackCount   = 1,
                                                TorEffectPredicateDelegate  effectQualifier = null)
        {
            Contract.Requires<ArgumentNullException>(onWhom != null);
            Contract.Requires<ArgumentException>(!string.IsNullOrEmpty(debuffName));

            effectQualifier = effectQualifier ?? (d => true);   // Resolve null effectQualifier to something sane

            string      normalizedDebuffName = Mirror.LookupBuff(debuffName);

            return (onWhom.Debuffs.FirstOrDefault(d => ((d.Name == normalizedDebuffName)
                                                                && effectQualifier(d)
                                                                && ((d.Caster.Name == BuddyTor.Me.Name) || string.IsNullOrEmpty(d.Caster.Name))    // ownership or shared
                                                                && (d.Stacks >= minStackCount)))  != null);
        }

    }
}
