﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using Buddy.BehaviorTree;
using Buddy.Common.Math;
using Buddy.CommonBot;
using Buddy.Swtor;

using Action = Buddy.BehaviorTree.Action;

namespace WingIt
{
    public static class Spell
    {
        /// <summary>
        /// <para>WaitForCast() for use in a behavior tree <c>PrioritySelector</c>.</para>
        /// <para>Returns <c>RunStatus.Success</c> while the spell cast is in progress; otherwise, returns <c>RunStatus.Failure</c>.</para>
        /// </summary>
        /// <returns>A <c>Composite</c> for use in a <c>PrioritySelector</c>.</returns>
        public static Composite WaitForCast()
        {
            return new Decorator(doWhen => BuddyTor.Me.IsCasting,
                new Action(ctx => RunStatus.Success));
        }

        /// <summary>
        /// <para>Casts ABILITYNAME, if REQUIREMENTS are met.</para>
        /// <para>For use in a behavior tree <c>PrioritySelector</c>.</para>
        /// <para>Returns <c>RunStatus.Success</c> if the ABILITYNAME was actually cast.  Otherwise, returns <c>RunStatus.Failure</c></para>
        /// <para>Failure to cast the spell can occur for a number of reasons:<list type="bullet"></list>
        /// <item><description><para> * the REQUIREMENTS are not met</para></description></item> 
        /// <item><description><para> * the toon has not learned the ability (or the ABILITYNAME is misspelled by the caller)</para></description></item> 
        /// <item><description><para> * the target is not within the spell's range</para></description></item> 
        /// <item><description><para> * the ability is on cooldown</para></description></item> 
        /// <para>Notes:</para>
        /// <item><description><para> * if REQUIREMENTS are omitted, ability use will be blindly attempted.</para></description></item> 
        /// <item><description><para> * the caller does not need to provide REQUIREMENTS for ability learned, distance to use, or cooldowns.
        /// All these checks are implicitly built into this method.</para></description></item> 
        /// </para>
        /// </summary>
        /// <param name="abilityName"></param>
        /// <param name="requirements"></param>
        /// <returns></returns>
        public static Composite Cast(string abilityName, BooleanValueDelegate requirements = null)
        {
            requirements = requirements ?? (ctx => true);   // Resolve a null argument into something sane

            // CombatRoutine.Cast() checks that we have learned the ability and
            // that the target is in range, so no need to duplicate that here...
            // (Verified by a conversation with Aevitas while he was looking at the BWcore source -- 28Jun2012-19:07UTC chinajade)
            return (WingIt.Instance.Cast(abilityName, requirements));
        }

        /// <summary>
        /// <para>Casts ABILITYNAME on ONCHARACTER, if REQUIREMENTS are met.</para>
        /// <para>For use in a behavior tree <c>PrioritySelector</c>.</para>
        /// <para>Returns <c>RunStatus.Success</c> if the ABILITYNAME was actually cast.  Otherwise, returns <c>RunStatus.Failure</c></para>
        /// <para>Failure to cast the spell can occur for a number of reasons:<list type="bullet"></list>
        /// <item><description><para> * the REQUIREMENTS are not met</para></description></item> 
        /// <item><description><para> * the toon has not learned the ability (or the ABILITYNAME is misspelled by the caller)</para></description></item> 
        /// <item><description><para> * the target is not within the spell's range</para></description></item> 
        /// <item><description><para> * the ability is on cooldown</para></description></item> 
        /// <para>Notes:</para>
        /// <item><description><para> * if REQUIREMENTS are omitted, ability use will be blindly attempted.</para></description></item> 
        /// <item><description><para> * the caller does not need to provide REQUIREMENTS for ability learned, distance to use, or cooldowns.
        /// All these checks are implicitly built into this method.</para></description></item> 
        /// </para>
        /// </summary>
        /// <param name="abilityName"></param>
        /// <param name="requirements"></param>
        /// <returns></returns>
        public static Composite Cast(string abilityName, CharacterSelectionDelegate onCharacter, BooleanValueDelegate requirements = null)
        {
            requirements = requirements ?? (ctx => true);   // Resolve a null argument into something sane

            // CombatRoutine.Cast() checks that we have learned the ability and
            // that the target is in range, so no need to duplicate that here...
            // (Verified by a conversation with Aevitas while he was looking at the BWcore source -- 28Jun2012-19:07UTC chinajade)
            return WingIt.Instance.Cast(abilityName, onCharacter, requirements);
        }

        /// <summary>
        /// <para>Casts ABILITYNAME on LOCATION, if REQUIREMENTS are met.</para>
        /// <para>For use in a behavior tree <c>PrioritySelector</c>.</para>
        /// <para>Returns <c>RunStatus.Success</c> if the ABILITYNAME was actually cast.  Otherwise, returns <c>RunStatus.Failure</c></para>
        /// <para>Failure to cast the spell can occur for a number of reasons:<list type="bullet"></list>
        /// <item><description><para> * the REQUIREMENTS are not met</para></description></item> 
        /// <item><description><para> * the toon has not learned the ability (or the ABILITYNAME is misspelled by the caller)</para></description></item> 
        /// <item><description><para> * the target location is not within the spell's range</para></description></item> 
        /// <item><description><para> * the ability is on cooldown</para></description></item> 
        /// <para>Notes:</para>
        /// <item><description><para> * if REQUIREMENTS are omitted, ability use will be blindly attempted.</para></description></item> 
        /// <item><description><para> * the caller does not need to provide REQUIREMENTS for ability learned, distance to use, or cooldowns.
        /// All these checks are implicitly built into this method.</para></description></item> 
        /// </para>
        /// </summary>
        /// <param name="abilityName"></param>
        /// <param name="requirements"></param>
        /// <returns></returns>
        public static Composite CastOnGround(string abilityName, BooleanValueDelegate requirements, CommonBehaviors.Retrieval<Vector3> location)
        {
            requirements = requirements ?? (ctx => true);   // Resolve a null argument into something sane

            return new Decorator(ret =>  requirements(ret) && location != null && location(ret) != Vector3.Zero,
                new Action(ret => AbilityManager.Cast(abilityName, location(ret))));
        }

        /// <summary>
        /// <para>Convenience method to cast BUFFNAME on <c>BuddyTor.Me</c>, but only if the buff does not currently exist on the toon, and REQUIREMENTS are met.</para>
        /// <para>For use in a behavior tree <c>PrioritySelector</c>.</para>
        /// <para>Returns <c>RunStatus.Success</c> if the BUFFNAME was actually cast.  Otherwise, returns <c>RunStatus.Failure</c></para>
        /// <para>Notes:
        /// <item><description><para> * if REQUIREMENTS are omitted, buffing will be blindly attempted if BUFFNAME is missing.</para></description></item> 
        /// <item><description><para> * the caller does not need to provide REQUIREMENTS for ability learned, distance to use, or cooldowns.
        /// All these checks are implicitly built into this method.</para></description></item> 
        /// <item><description><para> * the name of the buff received, must also be the same as the name of the spell used to cast the buff, BUFFNAME.</para>
        /// <para>For instance, casting "Uber" must yield a buff called "Uber", not "Uber (Force)" or "Uber (Tech)".</para>
        /// <para>If such a disparity exists, you must use a normal Cast() with the appropriate requirements, instead.</para></description></item> 
        /// </para>
        /// </summary>
        /// <param name="buffName"></param>
        /// <param name="requirements"></param>
        /// <returns></returns>
        public static Composite BuffSelf(string buffName, BooleanValueDelegate requirements = null)
        {
            requirements = requirements ?? (ctx => true);   // Resolve a null argument into something sane

            return (Cast(buffName, onUnit => BuddyTor.Me, castWhen => !BuddyTor.Me.HasBuff(buffName) && requirements(castWhen)));
        }
    }
}
