﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Windows;

using Buddy.BehaviorTree;
using Buddy.Common;
using Buddy.CommonBot;
using Buddy.Swtor;
using Buddy.Swtor.Objects;

using WingIt.Dynamics;
using Action = Buddy.BehaviorTree.Action;

namespace WingIt
{
    public class WingIt : CombatRoutine
    {
        public int number { get { return 0; } }
        private Composite _ooc;
        private Composite _pull;
        private Composite _combat;

        public static CombatRoutine Instance { get; set; }

        #region Overrides of CombatRoutine

        public override string Name { get { return "WingIt - All In One"; } }

        public override Window ConfigWindow { get { return null; } }

        public override CharacterClass Class { get { return BuddyTor.Me.Class; } }
        public AdvancedClass AdvClass { get { return BuddyTor.Me.AdvancedClass; } }
        public SkillTreeId charSpec { get { return SpecHandler.GetSpec(); } }

        public override Composite OutOfCombat { get { return _ooc; } }

        public override Composite Pull { get { return _pull; } }

        public override Composite Combat { get { return _combat; } }

        public override void Dispose()
        {
            // empty (for now)
        }

        public override void Initialize()
        {
            Logger.Write("Level: " + BuddyTor.Me.Level);
            Logger.Write("Class: " + Class);
            Logger.Write("Advanced Class: " + AdvClass + " (" + SpecHandler.GetSpec().ToString().Replace(AdvClass.ToString(), "") + ")");
            Instance = this;
            _ooc = new PrioritySelector(
                        CreateBT_CompanionHandler(),
                        CreateBT_ToonAndCompanionRejuvenate(),
                        CreateBT_ToonSelfBuff()
                        );

            int count;

            var extraOoc = CompositeBuilder.GetComposite(Class, AdvClass, charSpec, BehaviorType.OutOfCombat, out count);
            // Add the extra logic for OOC stuff from the combat declaration.
            if (extraOoc != null && count > 0)
                (_ooc as PrioritySelector).AddChild(extraOoc);

            _combat = CompositeBuilder.GetComposite(Class, AdvClass, charSpec, BehaviorType.Combat, out count);
            if (count == 0 || _combat == null)
            {
                Logger.Write("Combat support for " + Class + " [" + charSpec + "] is not currently implemented.");
                BotMain.Stop();
            }

            _pull = CompositeBuilder.GetComposite(Class, AdvClass, charSpec, BehaviorType.Pull, out count);
            if (count == 0 || _pull == null)
            {
                Logger.Write("Pull support for " + Class + " is not currently implemented.");
                if (_combat != null)
                {
                    _pull = _combat;
                    Logger.Write("Using Pull for combat.");
                }
                else
                {
                    BotMain.Stop();
                    Logger.Write("Buddy Wing stopped!");
                }
            }
        }

        #endregion      // Overrides of CombatRoutine


        private Composite CreateBT_CompanionHandler()
        {
            return new PrioritySelector(
               new Decorator(ctx => Helpers.IsCompanionInUse() && Helpers.Companion.IsDead,
                    new PrioritySelector(
                        CommonSubBT.CreateBTPS_BreakCrouchOrCover(),
                        new Decorator(ctx => (Helpers.Companion.Distance > 0.2f),
                            Movement.MoveTo(ctx =>Helpers.Companion.Position, 0.2f)),
                        CommonSubBT.CreateBTPS_StopMoving(),

                        // Otherwise, kick off the cast...
                        // NB: This should really be split into two pieces: one that kicks off the Revive, and another part
                        // that waits for the Revive to complete.  However, some BT node above BehaviorType.OutOfCombat is
                        // misbehaving.   Once casting starts, the upper-level node is waiting for (full term) casting to complete,
                        // thus depriving us of the chance to terminate the cast when it is no longer needed.
                        // C'est la vie.  We work around it by bundling the cast with a corresponding WaitContinue.
                        // Clumsy, but it deprives the upper node of the chance to misbehave.  Our workaround does not hinder
                        // higher-level nodes (such as combat) from running, if needed.
                        new Decorator(ctx => !BuddyTor.Me.IsCasting,
                            new Sequence(
                                Spell.Cast("Revive Companion", on => Helpers.Companion, when => true),
                                // Allow the SWtOR client/server time to update IsCasting from our Revive casting action...
                                new WaitContinue(TimeSpan.FromMilliseconds(1000), abortWaitingWhen => BuddyTor.Me.IsCasting, new Action()),
                                // Wait for Revive to actually complete...
                                new WaitContinue(TimeSpan.FromMilliseconds(15000), abortWaitingWhen => !BuddyTor.Me.IsCasting, new Action())
                                ))
                        ))

                // TODO: Worry about summon when the BuddyWing API supports it...
                //new Decorator(ctx => BuddyTor.Me.CompanionSummonable, //Summon)
               );
        }

        private Composite   CreateBT_ToonAndCompanionRejuvenate()
        {
            return new Decorator(ctx => IsMeOrCompanionRejuvenationNeeded(),
                    new PrioritySelector(rejuvenateAbilityName => Tunables.ForToon.RejuvenateName,
                        CommonSubBT.CreateBTPS_BreakCrouchOrCover(),
                        CommonSubBT.CreateBTPS_StopMoving(),

                        // Otherwise, kick off the cast...
                        // NB: This should really be split into two pieces: one that kicks off the heal, and another part
                        // that waits for the heal to complete.  However, some BT node above BehaviorType.OutOfCombat is
                        // misbehaving.   Once casting starts, the upper-level node is waiting for (full term) casting to complete,
                        // thus depriving us of the chance to terminate the cast when it is no longer needed.
                        // C'est la vie.  We work around it by bundling the cast with a corresponding WaitContinue.
                        // Clumsy, but it deprives the upper node of the chance to misbehave.  Our workaround does not hinder
                        // higher-level nodes (such as combat) from running, if needed.
                        new Decorator(rejuvenateAbilityName => !BuddyTor.Me.HasBuff((string)rejuvenateAbilityName),
                            new Sequence(
                                Spell.Cast(Tunables.ForToon.RejuvenateName, on => BuddyTor.Me, when => true),
                                // Allow the SWtOR client/server time to update HasBuff() and IsCasting from our rejuv casting action...
                                new WaitContinue(TimeSpan.FromMilliseconds(1000), abortWaitingWhen => BuddyTor.Me.IsCasting, new Action()),
                                // Wait for heal to actually complete...
                                new WaitContinue(TimeSpan.FromMilliseconds(15000),
                                                 rejuvenateAbilityName => !BuddyTor.Me.HasBuff((string)rejuvenateAbilityName)
                                                                          || IsMeAndCompanionRejuvenationComplete(),
                                                 new Action()),
                                // We may have casting-time left when fully healed, terminate the remainder of the 'channeling'...
                                CommonSubBT.CreateBTPS_StopCasting()
                                ))
                        ));
        }

        private Composite   CreateBT_ToonSelfBuff()
        {
            return new PrioritySelector(
                // NB: Currently, these buffs must be insta-cast, as the toon may be moving when they're cast...
                // If you need to violate this, you'll need to include a CommonBehaviors.MoveStop()
                // in your action, and use WaitContinue() for casting to complete.
                BuffSelf("Sprint"),
                BuffSelf(Tunables.ForToon.SelfBuffName)
                );
        }

        private bool IsMeAndCompanionRejuvenationComplete()
        {
            return (Tunables.ForToon.IsRejuvenationComplete() && Tunables.ForCompanion.IsRejuvenationComplete());
        }

        private bool IsMeOrCompanionRejuvenationNeeded()
        {
            return (Tunables.ForToon.IsRejuvenationNeeded() || Tunables.ForCompanion.IsRejuvenationNeeded());
        }
    }
}